/* a simple chat app using allnet */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/select.h>
/*
#include <readline/readline.h>
#include <readline/history.h>
*/

#include "types.h"
#include "messages.h"
#include "api.h"

/* maximum string size for a variety of things */
#define MAGIC_SIZE	141

static char prompt [MAGIC_SIZE];
static int waiting_for_ack = 0;
static char sent_message_owner_id [OWNER_ID_SIZE];
static char expecting_peer [MAGIC_SIZE];
static char expecting_secret [MAGIC_SIZE];
static char peer_shared_secret [MAGIC_SIZE];
static int received_key_count = -1;
/* while initializing don't print, print messages later in reverse order */
static int initializing = 1;
#define MAX_SAVED_MESSAGES	1000
struct message_info {
  char string [MAGIC_SIZE];
  time_t gtime;
  int counter;
  int key_id;
};
static struct message_info messages [MAX_SAVED_MESSAGES];
static int saved_messages = 0;

static pthread_cond_t ack_cond = PTHREAD_COND_INITIALIZER;

static void strip_non_alpha (char * s)
{
  char * original = s;
  while (*s != '\0') {
    if (! isalpha (*s)) {
      char * from = s + 1;
      char * to = s;
      while (*from != '\0')
        *(to++) = *(from++);
      *to = '\0';
    } else {
      s++;
    }
  }
}

static void make_upper (char * s)
{
  while (*s != '\0') {
    *s = toupper (*s);
    s++;
  }
}

static void make_lower (char * s)
{
  while (*s != '\0') {
    *s = tolower (*s);
    s++;
  }
}

static int local_time_offset (struct tm * ltime_tm, time_t * gtime)
{
  struct tm gtime_tm;
  gmtime_r (gtime, &gtime_tm);
  time_t ltime = mktime (ltime_tm);
  time_t gtime2 = mktime (&gtime_tm);
/*
  printf ("gtime is %ld = %s, gtime 2 is %ld\n",
          *gtime, asctime (&gtime_tm), gtime2);
  printf ("ltime is %ld = %d:%02d = %s\n", ltime,
          ltime_tm->tm_hour, ltime_tm->tm_min, asctime (ltime_tm));
*/
  /* printf ("local time offset is %ld\n", (ltime - gtime2) / 60); */
  return (ltime - gtime2) / 60;
}

void time_to_string (char * result, int rsize, struct tm * my_ltime_tm,
                     int sender_time_offset, int my_time_offset)
{
  int offset = 26;
  if (rsize <= offset) {
    snprintf (result, rsize, "too small!");
    return;
  }
  asctime_r (my_ltime_tm, result);
  * (rindex (result, '\n')) = '\0';
  offset = strlen (result);
  if (sender_time_offset == my_time_offset) /* easy case, we are finished */
    return;
  int delta_min = (sender_time_offset - my_time_offset) % 60;
  int delta_hour = (sender_time_offset - my_time_offset) / 60;
  /* printf ("offset is %d - %d = %02d:%02d\n",
          sender_time_offset, my_time_offset, delta_hour, delta_min); */
  if (delta_min == 0) {
    if (delta_hour == 1)
      snprintf (result + offset, rsize - offset, " (+1 hour at sender)");
    else if (delta_hour == -1)
      snprintf (result + offset, rsize - offset, " (-1 hour at sender)");
    else
      snprintf (result + offset, rsize - offset, " (%+d hours at sender)",
                delta_hour);
  } else {
    snprintf (result + offset, rsize - offset, " (%+d:%02d at sender)",
              delta_hour, delta_min);
  }
}

static void send_key_hash (int port, char * shared_secret)
{
  int mlen;
  char * message = message_for_key_exchange (shared_secret, &mlen);
  /* broadcast in the clear, without any owner ID */
  allnet_send_clear (port, message, mlen, NULL, 0, 0);
  free (message);
}

/* in case of error, returns NULL.  Otherwise, returnes a malloc'd string
 * with the user input.  Never returns an empty string, that is, a string
 * with the first character set to '\0'.
 *
 * if finished is NULL, only returns after it gets a line.
 * if finished is not NULL, returns either after it gets a line, or
 * when *finished is -1, whichever happens first.  In the latter case,
 * returns NULL.
 *
 * if finished and message are not NULL, prints message if it returns
 * because *finished was set to -1
 */
static char * my_readline (char * prompt, int * finished, char * message)
{
  int alloc_size = 5;
  char * result = malloc (alloc_size);
  *result = '\0';
  printf ("%s", prompt);
  char * p = result;
  int nc = 0;
  while (1) {
    fd_set read_fds;
    FD_ZERO (&read_fds);
    FD_SET (STDIN_FILENO, &read_fds);
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10000;  /* 10 ms, for fast response */
    struct timeval *tp = &t;
    if (finished == NULL)   /* nothing to time out for */
      tp = NULL;
    if (select (STDIN_FILENO + 1, &read_fds, NULL, NULL, tp) == 0) {
       /* timeout, check to see if we are finished */
      if ((finished != NULL) && (*finished == -1)) {
        if (message != NULL)
          printf ("\n%s\n", message);
        free (result);
        return NULL;
      }
      continue;  /* do not read, restart executing the while loop */
    }
    if (read (STDIN_FILENO, p, 1) != 1) {
      perror ("read/stdin");
      free (result);
      return NULL;
    }
    if (*p == '\n') {
      *p = '\0';
      if (p == result) {
        free (result);
        return NULL;
      }  /* else result is not empty */
      /* printf ("my_readline returning '%s'\n", result); */
      return result;
    }
    p++;
    *p = '\0';
    nc++;
    if (nc + 1 > alloc_size) {
      alloc_size = alloc_size * 2 + 5;
      result = realloc (result, alloc_size);
      p = result + nc;
    }
  }
}

static void save_message (char * string, time_t * gtime, int counter, int key)
{
  int i;
  for (i = 0; i < saved_messages; i++) {
    if ((key == messages [i].key_id) &&
        (*gtime == messages [i].gtime) &&
        (counter == messages [i].counter) &&
        (strcmp (string, messages [i].string) == 0)) {
      printf ("message already saved\n");
      return;
    }
  }
  if (saved_messages < MAX_SAVED_MESSAGES) {
    strncpy (messages [saved_messages].string, string, MAGIC_SIZE);
    messages [i].gtime = *gtime;
    messages [i].counter = counter;
    messages [i].key_id = key;
    saved_messages++;
  } else {
    printf ("too many messages, unable to save %s\n", string);
  }
}

struct message_header {
  char time [8];  /* 6 bytes of seconds since Jan 1st, 1970 GMT,
                   * 2 bytes of +- minute offset from GMT,
                   * offset 0x8000 if not known and local time is sent
                   */
  int counter;    /* each subsequent message has counter increased by 1,
                   * starting with 0 for the first message */
};

static void handle_user_message (int port, char * buffer, int bsize,
                                 struct allnet_addr * route, int route_num_addr,
                                 int route_used, int key_id)
{
/*
  printf ("received user message: ");
  print_buffer (buffer, bsize, 1);
*/

  int offset = OWNER_ID_SIZE + sizeof (struct message_header);
  if (bsize < offset) {
    printf ("minimum message size %d, received %d\n", offset, bsize);
    return;
  }

/*
  printf ("handle_user_message (%d, %p, %d, ... %d)\n",
          port, buffer, bsize, key_id);
*/
  int asize;
  char * ack = allnet_ack (buffer, bsize, &asize);
/* this simply puts my address back in, which is useless -- fix this
 * once allnetd actually records the route.
  reverse_route (route, route_num_addr, &route_used, route);
 */
  route_num_addr = 1;
  route_used = 0;
  struct allnet_addr dest;
  dest.discrim = ALLNET_BCAST_ADDR_ID;
  dest.bc_addr.hops = 6;
  allnet_send_clear (port, ack, asize, &dest, route_num_addr, route_used);
  free (ack);

  char * headerp = buffer + OWNER_ID_SIZE;
/*
  printf ("received header: ");
  print_buffer (headerp, 12, 1);
*/
  /* using long long otherwise compiler complains on systems with 32-bit time */
  long long gtimel = (((long long) (headerp [0] & 0xff)) << 40) |
                     (((long long) (headerp [1] & 0xff)) << 32) |
                     (((long long) (headerp [2] & 0xff)) << 24) |
                     (((long long) (headerp [3] & 0xff)) << 16) |
                     (((long long) (headerp [4] & 0xff)) <<  8) |
                      ((long long) (headerp [5] & 0xff));
  time_t sender_gtime = gtimel;
  /* headerp [6] is not & 0xff because here we want sign extension, i.e.
   * for negative values to turn into negative integers */
  int sender_time_offset = ((headerp [6]) << 8) | (headerp [7] & 0xff);
  int counter = ((headerp  [8] & 0xff) << 24) |
                ((headerp  [9] & 0xff) << 16) |
                ((headerp [10] & 0xff) <<  8) |
                 (headerp [11] & 0xff);
  struct tm my_ltime_tm;
  localtime_r (&sender_gtime, &my_ltime_tm);
  /* int my_time_offset = local_time_offset (&my_ltime_tm, &sender_gtime_tm); */
  int my_time_offset = local_time_offset (&my_ltime_tm, &sender_gtime);

  char string [MAGIC_SIZE];
  int size = MAGIC_SIZE - 1;
  if (bsize - offset < size)
    size = bsize - offset;
  memcpy (string, buffer + offset, size);
  string [size] = '\0';

/*
  printf ("received message sent at time %ld, offset %d, now is %ld\n",
          sender_gtime, sender_time_offset, time (NULL));
*/
  char time_string [MAGIC_SIZE];
  time_to_string (time_string, MAGIC_SIZE, &my_ltime_tm,
                  sender_time_offset, my_time_offset);

  char name [MAGIC_SIZE] = "unknown";
  char user_path [MAGIC_SIZE] = "";
  name_for_key (key_id, name, MAGIC_SIZE - 1, user_path, MAGIC_SIZE - 1);

  char to_print [MAGIC_SIZE * 3];
  snprintf (to_print, sizeof (to_print),
            "(%d) from %s at %s: %s", counter, name, time_string, string);

  if (! initializing)
    printf ("\n  %s\n%s", to_print, prompt);
  else
    save_message (to_print, &sender_gtime, counter, key_id);
}

static void handle_clear_message (int port, char * buffer, int bsize)
{
  if ((waiting_for_ack) &&
      (is_valid_ack (buffer, bsize, sent_message_owner_id, OWNER_ID_SIZE))) {
    waiting_for_ack = 0;
    pthread_cond_broadcast (&ack_cond);
    return;
  /* else, may be a key exchange message */
  } else if ((received_key_count == -1) ||  /* not expecting anything */
             (received_key_count > 100)) {  /* too many tries */
    if (received_key_count != -1)
      printf ("received %d keys, ignoring this one\n", received_key_count);
/*
    else
      printf ("\nz: %s", prompt);
*/
    received_key_count = -1;
  } else {
    int verification = verify_key_exchange (buffer, bsize, expecting_secret,
                                            expecting_peer);
    if (verification >= 0) {
      printf ("\nreceived and saved key for %s\n", expecting_peer);
      received_key_count = -1;
      send_key_hash (port, expecting_secret);
    } else {
      if (verification == -2) {
        /* printf ("received my own key message, ignoring\n"); */
      } else {
        printf ("message of length %d, verification %d\n", bsize, verification);
        received_key_count++;
      }
    }
  }
}

void receive_fun (void * arg, int port, char * buffer, int bsize,
                  struct allnet_addr * route, int route_num_addr,
                  int route_used, int key_id)
{
  /* printf ("received packet on port %d, key id %d\n", port, key_id); */
  if ((route_num_addr <= 0) || (route_used < 0)) { /* some error */
    printf ("receive_fun error: num_addrs %d, used %d\n", route_num_addr,
            route_used);
    exit (1);
  }
  if (key_id >= 0) {
    handle_user_message (port, buffer, bsize,
                         route, route_num_addr, route_used, key_id);
  } else {  /* key ID is zero, message is in the clear, a key or an ack */
    handle_clear_message (port, buffer, bsize);
  }
  /* fflush (stdout); */
}

static void send_message (int port, char * user_input,
                          char * user, int current_key, int counter)
{
  if ((current_key == -1) || (counter == -1)) {
    printf ("no user selected, unable to send message\n");
  }
  char actual_message [MAGIC_SIZE * 3];
  time_t now_time;
  time (&now_time);
  struct tm ltime;
  localtime_r (&now_time, &ltime);
  int lt_offset = local_time_offset (&ltime, &now_time);
/*
  printf ("sending at time %ld, offset %d\n", now_time, lt_offset);
*/
  /* debugging */
  
  long long now_timel = now_time;
/* big-endian encoding of time, offset, and counter */
  actual_message  [0] = (now_timel >> 40) & 0xff;
  actual_message  [1] = (now_timel >> 32) & 0xff;
  actual_message  [2] = (now_timel >> 24) & 0xff;
  actual_message  [3] = (now_timel >> 16) & 0xff;
  actual_message  [4] = (now_timel >>  8) & 0xff;
  actual_message  [5] =  now_timel        & 0xff;

  actual_message  [6] = (lt_offset >> 8) & 0xff;
  actual_message  [7] =  lt_offset       & 0xff;

  actual_message  [8] = (counter  >> 24) & 0xff;
  actual_message  [9] = (counter  >> 16) & 0xff;
  actual_message [10] = (counter  >>  8) & 0xff;
  actual_message [11] =  counter         & 0xff;

  int offset = sizeof (struct message_header);

  struct tm * now = localtime (&now_time);
  int send_len = 0;
  send_len = snprintf (actual_message + offset,
                       sizeof (actual_message) - offset, "%s", user_input);
  if (send_len >= 0)
    send_len += offset;
  random_fill (sent_message_owner_id, sizeof (sent_message_owner_id), 1, 0);
  waiting_for_ack = 1;
  /* only really interested in the signal, not the mutex */
  pthread_mutex_t useless_mutex = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock (&useless_mutex);
  struct timespec finish_time;
  clock_gettime (CLOCK_REALTIME, &finish_time);
  finish_time.tv_sec += 3;   /* wait up to 3s for ack */

/*
  printf ("sending header: ");
  print_buffer (actual_message, 12, 1);
*/

/*
  printf ("encrypting buffer: ");
  print_buffer (actual_message, send_len, 1);
*/
  struct allnet_addr dest;
  dest.discrim = ALLNET_KEY_ADDR_ID;
  dest.pk_addr.hops = 6;
  dest.pk_addr.num_bits = 8;  /* may change later */
  int osize = sizeof (sent_message_owner_id);
  int bsize;
  char * buf = allnet_encrypt (actual_message, send_len, current_key,
                               sent_message_owner_id, osize, &bsize,
                               b (dest.pk_addr.num_bits), dest.pk_addr.pubkey);
  allnet_send (port, buf, bsize, &dest, 1, 0, NULL, 0);

  if (pthread_cond_timedwait (&ack_cond, &useless_mutex, &finish_time) == 0)
    pthread_mutex_unlock (&useless_mutex);
  if (waiting_for_ack == 0) {
    printf ("  (acknowledged)\n");
  } else { /* not acked, re-send it to whoever will store it */
    allnet_send (port, buf, bsize, &dest, 1, 0, sent_message_owner_id, osize);
  }
  waiting_for_ack = 0;
  free (buf);
}

static void send_key (int port)
{
  char p [MAGIC_SIZE];
  snprintf (p, sizeof (p), "enter the secret string of your new contact: ");
  char * shared_secret = my_readline (p, &received_key_count, "key received");
  if (shared_secret == NULL) {
    /* printf ("no string entered\n"); */
    return;
  }
  if ((*shared_secret == '\0') || (*shared_secret == '0')) {
    /* printf ("no string entered\n"); */
    free (shared_secret);
    return;
  }
  strip_non_alpha (shared_secret);
  make_upper (shared_secret);
  strncpy (expecting_secret, shared_secret, MAGIC_SIZE);

  send_key_hash (port, shared_secret);
}

static void receive_key (int port)
{
  int have_name = 0;
  char * peer_name =
    my_readline ("enter the name to use for your new contact: ", NULL, NULL);
  if ((peer_name == NULL) || (*peer_name == '\0')) {
    time_t now = time (NULL);
    snprintf (expecting_peer, MAGIC_SIZE, "key received on %s", ctime (&now));
    if (expecting_peer [strlen (expecting_peer) - 1] == '\n')
      expecting_peer [strlen (expecting_peer) - 1] = '\0';
  } else {
    strncpy (expecting_peer, peer_name, MAGIC_SIZE);
    have_name = 1;
  }
  if (peer_name != NULL)
    free (peer_name);
  received_key_count = 0;  /* get ready to receive */

  char * use_name = "your new contact";
  if (have_name)
    use_name = expecting_peer;

  new_keypair (use_name);

#define SECRET_LENGTH	(12 + 2)
  char random_secret [MAGIC_SIZE];
  random_fill (random_secret, SECRET_LENGTH, 1, 1);
  random_secret [SECRET_LENGTH / 3] = ' ';
  random_secret [2 * SECRET_LENGTH / 3] = ' ';
  random_secret [SECRET_LENGTH] = '\0';
#undef SECRET_LENGTH
  char lower_secret [MAGIC_SIZE];
  strcpy (lower_secret, random_secret);
  strcpy (expecting_secret, random_secret);
  strip_non_alpha (expecting_secret);
  make_upper (expecting_secret);
  make_lower (lower_secret);

  printf ("either you or %s must enter a secret string\n", use_name);
  printf ("if you enter it, enter the one shown on the screen of %s\n",
          use_name);
  printf ("if %s enters it, tell them to enter '%s' (or %s)\n",
          use_name, expecting_secret, lower_secret);
  send_key (port);
}

static void my_exit (int port)
{
  allnet_stop (port);
  printf ("goodbye\n");
  exit (0);
}

#define MENU_EXIT		0
#define MENU_NO_ACTION		1
#define MENU_NEW_USER		2
#define MENU_MESSAGE		7
#define MENU_EXCHANGE_KEY	9
static int main_menu (char * user, /* pointer to MAGIC_SIZE bytes */
                      char ** message) /* set if returning MENU_MESSAGE
                                          or MENU_NEW_USER */
{
  if (*user == '\0')
    snprintf (prompt, sizeof (prompt),
              "enter 0 to quit, 9 to exchange keys, or user name: ");
  else
    snprintf (prompt, sizeof (prompt),
              "0 quit, 9 exchange keys, or message for %s: ", user);
/*
    snprintf (prompt, sizeof (prompt),
              "0 quit, 1 new user, 9 exchange keys, or message for %s: ", user);
*/
  char * input = my_readline (prompt, NULL, NULL);
  if (input == NULL)
    return MENU_EXIT;
  if (strcmp (input, "0") == 0) {
    free (input);
    return MENU_EXIT;
  }
  if (strcmp (input, "9") == 0) {
    free (input);
    return MENU_EXCHANGE_KEY;
  }
  if (*user == '\0') {
    if (input == NULL)
      return MENU_NO_ACTION;
    if (*input == '\0') {
      free (input);
      return MENU_NO_ACTION;
    }
    if (*input == '0') {
      free (input);
      return MENU_EXIT;
    }
    /* name entered */
    *message = input;
    return MENU_NEW_USER;
  } /* else it is a message for user */
  *message = input;
  return MENU_MESSAGE;
}

static void set_user (char * new_name, int * key, int * counter,
                      char * nick, int nsize, char * user_name, int usize,
                      char * user_path, int upsize)
{
  if (key_for_name (new_name, key, 1) <= 0) {
    printf ("user %s has no key\n", new_name);
    *key = -1;
    *counter = -1;
    *nick = '\0';
  } else {
    strncpy (nick, new_name, nsize);
    *counter = 0;  /* to do: save and retrieve counter */
    if (name_for_key (*key, user_name, usize, user_path, upsize) > 0) {
      printf ("talking to user %s, key ID %d\n", user_name, *key);
    } else {  /* got key from key_for_name, it should be valid */
      printf ("error: name_for_key failed for key ID %d, nickname %s\n",
              *key, nick);
      exit (1);  /* figure out what happened */
    }
  }
}

static void usage (char * programname)
{
  printf ("usage: %s [-c configuration-directory] name\n", programname);
  printf ("  for example, to talk to Bill: %s Bill\n", programname);
  printf ("  to use a configuration directory named 'test': %s -c test Bill\n",
          programname);
  printf ("  if the name is missing, it can be provided later\n");
  exit (1);
}

main (int argc, char ** argv)
{
  /* should be done first, before any use of stdout */
  if (setvbuf (stdout, NULL, _IONBF, 0) != 0) {
    perror ("setvbuf");
    printf ("unable to set stdout to unbuffered mode\n");
  }

  char * home = getenv ("HOME");
  char path [MAGIC_SIZE];
  snprintf (path, MAGIC_SIZE, "%s/%s", home, ALLNET_API_DIRECTORY);
  if ((argc > 2) && (strcmp (argv [1], "-c") == 0)) {
    strncpy (path, argv [2], MAGIC_SIZE);
    argc -= 2;
    argv += 2;
  }

  int port = allnet_start (receive_fun, NULL);

  int num_keys = init_key_files (path, "chat");
  if (num_keys < 0) {
    printf ("unable to open key files, aborting\n");
    exit (1);
  }

  struct allnet_addr route [2];
  route [0].discrim = ALLNET_KEY_ADDR_ID;
  route [0].pk_addr.hops = 1;
  route [0].pk_addr.num_bits = 16;
  route [1].discrim = ALLNET_BCAST_ADDR_ID;
  route [1].bc_addr.hops = 6;
  /* request any outstanding messages */
  /* should put my key here, send with a pub key source address TO
   * a broadcast address. */
  /* note: messages are returned in priority order, meaning most recent
   * first.  Really should display oldest first */
  int old_saved_messages;
  do {
    old_saved_messages = saved_messages;
    int i;
/* does not scale, fix later */
    for (i = 0; i < num_keys; i++) {
      key_bytes (i, route [0].pk_addr.pubkey, 2);
      allnet_send_clear (port, REQUEST_BYTES, CLEAR_ID_SIZE, route, 2, 1);
    }
    sleep (1);
    if (old_saved_messages != saved_messages)
      printf ("got %d saved messages\n", saved_messages);
  } while (old_saved_messages != saved_messages);
  initializing = 0;
  int i, ik;
  for (ik = 0; ik < num_keys; ik++) {
    for (i = saved_messages - 1; i >= 0; i--) {
      if (messages [i].key_id == ik) {
        printf ("%s\n", messages [i].string);
      }
    }
  }

  char nick [MAGIC_SIZE] = "";
  char user_name [MAGIC_SIZE] = "";
  char user_path [MAGIC_SIZE] = "";
  int current_key = -1;
  int counter = -1;
  if (argc > 1)
    set_user (argv [1], &current_key, &counter, nick, sizeof (nick),
              user_name, sizeof (user_name), user_path, sizeof (user_path));
  while (1) {
    char * message;
    int request = main_menu (nick, &message);
    switch (request) {
      case MENU_EXIT:
        my_exit (port);
        break;             /* in case my_exit ever returns */
      case MENU_NO_ACTION:
        break;
      case MENU_NEW_USER:  /* user should have been set, set key */
        set_user (message, &current_key, &counter, nick, sizeof (nick),
                  user_name, sizeof (user_name), user_path, sizeof (user_path));
        break;
      case MENU_MESSAGE:
        send_message (port, message, nick, current_key, counter);
        counter++;
        break;
      case MENU_EXCHANGE_KEY:
        receive_key (port);
        break;
      default:
        printf ("unimplemented menu item %d\n", request);
        break;
    }
  }
}
